/*********************************************************************
 *
 *                Microchip C18 Firmware - RS-485 network
 *
 *********************************************************************
 * FileName:        user_serial.c
 * Dependencies:    see INCLUDES section below
 * Processor:       PIC18F4520
 * Compiler:        C18 
 * Company:         Lakeview Research LLC (www.Lvr.com)
 *
 * Software License Agreement

 * Licensor grants any person obtaining a copy of this software ("You") 
 * a worldwide, royalty-free, non-exclusive license, for the duration of 
 * the copyright, free of charge, to store and execute the Software in a 
 * computer system and to incorporate the Software or any portion of it 
 * in computer programs You write.   

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * Author               Date        Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Jan Axelson           10/19/07     
 *********************************************************************/

/*

The PIC is a secondary node in an RS-485 network. The USART interfaces 
to an RS-485 transceiver. A network can have multiple secondary nodes.

The PIC responds to defined commands. Each command has this format:
byte 1: ":"
byte 2: the PIC's network address
byte 3: command code
bytes 4..n: additional data (command-dependent)
final byte: LF code (0ah)

Each response has this format:
byte 1: ":"
byte 2: the PIC's network address
byte 3..n: additional data (command dependent)
final byte: LF code (0ah)

This example defines two commands, read_byte and write_byte.

To read a byte, send this command to the PIC:

:h2b 

followed by a LF code (hit Enter) 

The PIC will respond with:

:hxx

followed by a LF, where xx is the hex value read.

To write a byte, send this command to the PIC:

:h1bxx 

followed by a LF (hit Enter), where xx is the hex value to write.
(For example, :h1bff or :h1b00) 

The PIC will respond with:

:h

followed by a LF.

A PC can communicate with the PIC using a custom application, or you can use
a terminal-emulator such as HyperTerminal if the PC's driver-enable line 
is controlled by hardware only.

On the PIC, firmware can control the RS-485 driver-enable line if needed. 

This application and host applications to communicate with the device are available
from www.Lvr.com.
*/

/** I N C L U D E S **********************************************************/

#include <p18cxxx.h>
#include <usart.h>
#include "ctype.h"
#include "timers.h"
#include "user_rs485.h"
#include "system\typedefs.h"

/** V A R I A B L E S ********************************************************/

#pragma udata

#define COMMAND_START 58

// Comment out this line if the remote computer doesn't require extra time to disable 
// its RS-485 driver before the PIC responds to a received command.

#define delay_before_responding

// Comment out this line if you don't require firmware control of the RS-485 
// driver-enable line.

#define firmware_driver_enable

// Use any spare output port bit for the driver enable, if needed.

#if defined(firmware_driver_enable)
	#define driver_enable  PORTBbits.RB3
#endif

const byte MAX_COMMAND_LENGTH = 5;

// The PIC's network address. Use any character. Every node must have a unique address.

const char MY_ADDRESS = 'h';

byte command_index;
char command_response[6];
char network_state = 'r';
char received_command[6];
byte response_index = 0;
unsigned char serial_in;

/** P R I V A T E  P R O T O T Y P E S ***************************************/

int ascii_hex_to_byte(char, char);
int byte_to_ascii_hex(unsigned char value_to_convert, char converted_value[]);
void check_response_delay(void);
void command_read_byte(void);
void command_write_byte(void);
void initialize_serial_buffers(void);
void prepare_to_respond(void);
void respond_to_command(void);
void serial_communications(void);
void start_response_delay_timer(void);
void transmit_serial_data(void);
void wait_for_data_to_transmit(void);

/** D E C L A R A T I O N S **************************************************/

#pragma code

/******************************************************************************
 * Function:        byte ascii_hex_to_byte 
 *
 * PreCondition:    None
 *
 * Input:           char upper_nibble
					char lower_nibble
 *
 * Output:          the ASCII hex value converted to a byte or -1 on error
 *
 * Side Effects:    None
 *
 * Overview:        Converts two ASCII hex characters to a byte.			   	           
 *
 * Note:            Characters A-F can be upper or lower case
 *****************************************************************************/

int ascii_hex_to_byte(char upper_nibble, char lower_nibble)
{
	byte return_value;
	
	// Convert each char into its value.

	if ((upper_nibble >= '0') & (upper_nibble <= '9'))
		upper_nibble -= 48;
	else
	{
		// Change upper-case A-F to lower case.
		if ((upper_nibble >= 'A') & (upper_nibble <= 'F'))
			upper_nibble = tolower(upper_nibble);

		if ((upper_nibble >= 'a') & (upper_nibble <= 'f'))
			upper_nibble -= 87;
		else
			upper_nibble = -1;
	}

	if ((lower_nibble >= '0') && (lower_nibble <= '9'))
		lower_nibble -= 48;
	else
	{
		// Change upper-case A-F to lower case.

		if ((lower_nibble >= 'A') & (lower_nibble <= 'F'))
			lower_nibble = tolower(lower_nibble);
		if ((lower_nibble >= 'a') && (lower_nibble <= 'f'))
			lower_nibble -= 87;
		else
			lower_nibble = -1;
		}
	if ((upper_nibble > -1) && (lower_nibble > -1))
		return_value = (upper_nibble << 4) + lower_nibble;
	else
		return_value = -1;
	
	return return_value;
}

/******************************************************************************
 * Function:        int byte_to_ascii_hex
 *
 * PreCondition:    None
 *
 * Input:           unsigned char value_to_convert
 *					char converted_value[]
 *
 * Output:          0 on success, -1 on error
 *					upper_nibble and lower_nibble contain the character codes
 *
 * Side Effects:    None
 *
 * Overview:        Converts a byte into two ASCII hex characters                 
 *
 * Note:            None
 *****************************************************************************/

int byte_to_ascii_hex(unsigned char value_to_convert, char converted_value[])
{
	int success = 0;
	char upper_nibble;
	char lower_nibble;

	// Get each hex digit and convert it to a character code.

	upper_nibble = (value_to_convert & 0xf0) >> 16;

	if ((upper_nibble >= 0) && (upper_nibble <= 9))
		upper_nibble += 48;
	else
	{
			if ((upper_nibble >= 10) && (upper_nibble <= 15))
				upper_nibble += 87;	
			else
				success = -1;
	}
	lower_nibble = value_to_convert & 0x0f;

	if ((lower_nibble >=0) && (lower_nibble <= 9))
		lower_nibble += 48;
	else
	{
		if ((lower_nibble >= 10) && (lower_nibble <= 15))
			lower_nibble += 87;
		else
			success = -1;
	}
	converted_value[0] = upper_nibble;
	converted_value[1] = lower_nibble;
	return success;
}

/******************************************************************************
 * Function:        void check_response_delay 
 *
 * PreCondition:    Called only if delay_before_responding is defined.
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Checks if the delay has timed out. If yes, set the network 
 *					state to enable transmitting a response.                  
 *
 * Note:            None
 *****************************************************************************/

void check_response_delay(void)
{
	#if defined(delay_before_responding)

		if (INTCONbits.TMR0IF == 1)
		{
			// The primary node's driver should be disabled. Stop the timer.
			// OK to send the response.

			CloseTimer0;
			INTCONbits.TMR0IF = 0;
			network_state = 't';
		}
	#endif
}

/******************************************************************************
 * Function:        command_read_byte
 *
 * PreCondition:    A command has been received with command value = '1'.
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Prepares to send a byte in a response to the primary node.                  
 *
 * Note:            Currently supports only reading address 'b', defined as
 *					"read Port B".
 *****************************************************************************/

void command_read_byte(void)
{
	if (received_command[3] == 'b')
	{
		// Read Port B. Store the value read and a LF to send in the response.

		byte_to_ascii_hex(PORTB, &command_response[2]);
		command_response[4] = 0x0a;

		prepare_to_respond();
	}				
}

/******************************************************************************
 * Function:        command_write_byte	 
 *
 * PreCondition:    A command has been received with command value = '2'.
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Processes a received byte and prepares a response.                  
 *
 * Note:            Currently supports only writing to address 'b', defined as
 *					"write received bit 0 to Port B, bit 0".
 *****************************************************************************/

void command_write_byte(void)
{
	int received_data;

	if (received_command[3] == 'b')
	{
		// Get the data to write.
	
		received_data = ascii_hex_to_byte(received_command[4], received_command[5]);
		
		if (received_data > -1)
		{		
			// The received data was valid.
			// Set a port bit to match bit 0 of the received byte.

			if ((received_data & 1) == 1)
			{
				PORTBbits.RB0 = 1;
			}
			else
			{	
				PORTBbits.RB0 = 0;
			}

			command_response[2] = 0x0a;
			prepare_to_respond();	
		}
	}
}

/******************************************************************************
 * Function:        configure_command_response 
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Store the first 2 bytes in a command response.
 *					These values never change.
 *
 * Note:            None
 *****************************************************************************/

void configure_command_response(void)
{	
	command_response[0] = COMMAND_START;
	command_response[1] = MY_ADDRESS;
}

/******************************************************************************
 * Function:        void configure_serial_port(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Configures the UART for asynchronous communications.
 *                  RX = Port C, bit 7 
 *					TX = Port C, bit 6
 *
 * Note:            None
 *****************************************************************************/

void configure_serial_port(void)
{
	command_index = 0;

	// The bit rate is determined by the last two parameters in OpenUSART,
	// BAUDCONbits.BRG16, and FOSC. See your chip's data sheet for details.

	// Assumes FOSC = 4 Mhz, 9600 bps.
	
	OpenUSART (USART_TX_INT_OFF & 
 			   USART_RX_INT_OFF &
	           USART_ASYNCH_MODE & 
			   USART_EIGHT_BIT &
			   USART_CONT_RX & 
		 	   USART_BRGH_HIGH, 
			   25);

	#if defined(firmware_driver_enable)
		TRISBbits.TRISB3 = 0;
		driver_enable = 0;
	#endif
}

/******************************************************************************
 * Function:        void initialize_serial_buffers
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Initializes the buffers that store received commands
 *					and responsed to commands.              
 *
 * Note:            None
 *****************************************************************************/

void initialize_serial_buffers(void)
{
	command_index = 0;
	response_index = 0;		
	received_command[0]= '\0';
}

/******************************************************************************
 * Function:        void prepare_to_respond
 *
 * PreCondition:    A response is ready to send.
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Initialize the response buffer and set the network state.
 *
 * Note:            None
 *****************************************************************************/

void prepare_to_respond(void)
{
	response_index = 0;
	#if defined(delay_before_responding)		
		start_response_delay_timer();
		network_state = 'd';
	#else
		network_state = 't';
	#endif
}

/******************************************************************************
 * Function:        void receive_serial_data(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Read a received character and take appropriate action.
 *
 * Note:            None
 *****************************************************************************/

void receive_serial_data(void)
{
  	if (DataRdyUSART())
	{
		// The serial port has received a byte.
		// Read a received character.

		if (RCSTAbits.OERR == 1)
		{
			// Overrun error. 
			// Clear the error and re-enable the receiver.
	
			RCSTAbits.CREN = 0;
			RCSTAbits.CREN = 1;
		}
		if (RCSTAbits.FERR == 1)
		{
			// Framing error.
			// Reading a byte clears the error.
			// Read the byte but don't use it.

			serial_in = getcUSART(); 
		}
		else
		{
			// A character was received without a framing error.
	
	    	serial_in = getcUSART(); 

			switch (serial_in)
			{
				case 0x0a:
				{
					// A LF character was received, indicating the end of a command.
				
					respond_to_command();
	
					break;
				}
				
				case 0x0d:
				{
					// Ignore a received CR character. 
				
					break;
				}

				case COMMAND_START:
				{							 
					// A new command has begun. 
					// Initialize the array that holds received characters. 
	
					received_command[0]=COMMAND_START;
					command_index = 1;
					break;
				}	
				default:
				{
					// A character was received and it's not a LF or CR.
					// Save the character and increment the position in the variable
					// that stores received text.
					// Convert the char to lower case to make the input case-insensitive.				

					// The primary node might send extraneous data.
					// If at the end of the array, ignore additional received data.
				
					if (command_index <= MAX_COMMAND_LENGTH)
					{
						received_command[command_index] = tolower(serial_in);
						command_index++;
					}
					break;
				}
			}
		}
	}
}	

/******************************************************************************
 * Function:        void respond_to_command(void)
 *
 * PreCondition:    A command was received.
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Find out if a received command matches a defined command.
 *					If yes, perform the appropriate action.                  
 *
 * Note:            None
 *****************************************************************************/

void respond_to_command(void) 
{
	int received_data;
	
	if (received_command[0] == COMMAND_START)
	{
		if (received_command[1] == MY_ADDRESS)
		{
			// Respond to supported commands.

			switch (received_command[2])
			{
				case '1':
				{
					command_write_byte();				
					break;
				}		
				case '2':
				{
					command_read_byte();			
					break;
				}
				default:

					// Wait for another command.

				 	initialize_serial_buffers;		
					network_state = 'r';			
					break;				
			}
		}
		else
		{
			// Wait for another command.

		 	initialize_serial_buffers;
			network_state = 'r';
		}
	}
	else
	{
		// Wait for another command.

		initialize_serial_buffers;
		network_state = 'r';
	}
}

/******************************************************************************
 * Function:        void serial_communications 
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        State machine for serial communications. The states are:
 *
 *					'r' - waiting to receive data
 *					'd' - delaying after receiving a command and before responding
 *					't' - transmitting a response                 
 *
 * Note:            State 'd' is used only if delay_before_responding is defined.
 *****************************************************************************/

void serial_communications(void)
{
	switch (network_state)
	{
		case 'r':
		{
			// Receive data.

			receive_serial_data();			
			break;
		}		
		case 'd':
		{
			// Find out if the delay time has elapsed.

			check_response_delay();				
			break;
		}
		case 't':
		{
			// Transmit data.

			transmit_serial_data();		
			break;
		}
		default:
			break;				
	}
}

/******************************************************************************
 * Function:        void start_response_delay_timer
 *
 * PreCondition:    Is called only if delay_before_responding is defined.
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Sets the delay between receiving a command and sending
 *       			a response. The delay time varies with the requirements
 *					of the remote computer. The timer settings also vary
 *					with FOSC.                  
 *
 * Note:            None
 *****************************************************************************/

void start_response_delay_timer(void)
{
	// This example sets a delay of around 0.5 sec. assuming FOSC = 4 Mhz.

	OpenTimer0( TIMER_INT_OFF &
				T0_16BIT &
				T0_SOURCE_INT &
				T0_PS_1_256 );
	WriteTimer0(0xf800);
}

/******************************************************************************
 * Function:        void transmit_serial_data
 *
 * PreCondition:    None
 *
 * Input:           
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        If a byte is ready to send, write it to the serial port. 
 *                  
 *
 * Note:            None
 *****************************************************************************/

void transmit_serial_data(void)
{
	#if defined firmware_driver_enable
		driver_enable = 1;
	#endif	
	
	while(BusyUSART());
	putcUSART (command_response[response_index]);
	
	if (command_response[response_index] == 0x0a)
	{
		// The entire response has been sent.
	
	
		#if defined(firmware_driver_enable)
			while(BusyUSART());  	
			driver_enable = 0;   
		#endif				
	
		// Ready to receive another command.
		// Reinitialize the command and response variables.

		initialize_serial_buffers;	
		network_state = 'r';
	}
	else
	{
		// Prepare to send the next byte.
	
		response_index++;
	}
}


